// 菊地文雄, 山本昌宏「微分方程式と計算機演習」, 山海堂 (1991), p. 125
/* **        EIGEN ANALYSIS OF       **
   ** ORDINARY DIFFERENTIAL EQUATION **
   N : NO．OF DIVISIONS
   SHIFT : SHIFT PARAMETER
   EPS : THRESHOLD VALUE FOR ERROR
   M : MAXIMUM NO. OF ITERATIONS
   EVAL : APPROXIMATE EIGENVALUE
   H : MESH SIZE
   A,B,C : COEFFICIENT MATRIX
   U : EIGENVECTOR
   NITER : NO. OF ITERATIONS
   NPIVOT : NO. OF NEGATIVE PIVOTS
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

void shinv(double a[], double b[], double c[],
       double x[], double y[],
       double eps, double h, double shift,
       int n, int m, double *eval, int *niter, int *npivot);
void trid(double a[], double b[], double c[], double f[], int n, int id);
void inpr(double x[], double y[], int n, double *s);
double rnd(int *i);

#define NDIM (100)

int main(void)
{
  int n, m, i, niter, npivot;
  double shift, eps, h, h2, eval;
  double a[NDIM], b[NDIM], c[NDIM], u[NDIM], v[NDIM];
  // INPUT
  printf("INPUT : N=\n");
  scanf("%d", &n);
  if (n <= 1) exit(1);
  printf("INPUT : SHIFT=\n");
  scanf("%lf", &shift);
  printf("INPUT : EPS=\n");
  scanf("%lf", &eps);
  printf("INPUT : M=\n");
  scanf("%d", &m);
  // CALCULATION OF COEFFICIENT MATRIX
  h = 1.0 / n;
  h2 = h * h;
  for (i=0; i<n-1; i++) {
    a[i] = -1.0 / h2;
    b[i] = 2.0 / h2 - shift;
    c[i] = -1.0 / h2;
  }
  // CALCULATION OF EIGENPAIR
  shinv(a,b,c,u,v,eps,h,shift,n-1,m,&eval,&niter,&npivot);
  // OUTPUT
  printf("NO. OF ITERATIONS=%3d   EPS=%12.4e   SHIFT=%12.4e   NPIVOT=%3d\n",
     niter, eps, shift, npivot);
  printf("EIGENVALUE=%12.4e\n", eval);
  printf("** NODAL VALUES OF EIGENFUNCTION **\n");
  for (i = 0; i < 5; i++)
    printf("  I     U(I)    ");
  printf("\n");
  for (i=0; i<n-1; i++) {
    printf("%3d%12.4e ", i+1, u[i]);
    if ((i+1) % 5 == 0 || (i+1) == n-1)
      printf("\n");
  }
}

void shinv(double a[], double b[], double c[],
       double x[], double y[],
       double eps, double h, double shift,
       int n, int m, double *eval, int *niter, int *npivot)
{
  int j, i;
  double s, t, enew, error;
  // STARTING QUANTITIES
  j=1;
  for (i=0; i<n; i++)
    x[i] = rnd(&j);
  *niter=0;
  *eval=shift;
  // ITERATION
  do {
    for (i=0; i<n; i++)
      y[i] = x[i];
    trid(a,b,c,y,n,*niter);
    inpr(x,y,n,&s);
    inpr(y,y,n,&t);
    enew=s/t+shift;
    t=sqrt(h*t);
    for (i=0; i<n; i++)
      x[i] = y[i] / t;
    error=fabs((enew-*eval)/enew);
    *eval=enew;
    (*niter)++;
    printf("(NITER,EVAL,ERROR)=%d %g %g\n", *niter, *eval, error);
    if (*niter == m) break;
  }
  while (error > eps);
  *npivot=0;
  for (i=0; i<n; i++)
    if (b[i] < 0.0) (*npivot)++;
}

double rnd(int *i)
{
  // PSEUDO-RANDOM NUMBERS
  *i = 12869 * (*i) + 6925;
  *i = *i % (1<<15);
  return  (double) *i / (1 << 15);
}

void trid(double a[], double b[], double c[], double f[], int n, int id)
{
  int i;
  double aa;
  // FORWARD ELIMINATION
  for (i=0; i<n-1; i++) {
    aa = a[i+1] / b[i];
    if (id == 0) b[i+1] -= aa * c[i];
    f[i+1] -= aa * f[i];
  }
  // BACKWARD SUBSTITUTION
  f[n-1] /= b[n-1];
  for (i=n-2; i>=0; i--)
    f[i] = (f[i] - c[i]*f[i+1]) / b[i];
}

void inpr(double x[], double y[], int n, double *s)
{
  int i;
  // INNER PRODUCT
  *s=0.0;
  for (i=0; i<n; i++)
    *s += x[i] * y[i];
}
